@@ -1047,7 +1047,7 @@ NS_IMETHODIMP
 nsCanvasRenderingContext2D::Fill()
 {
     ApplyStyle(STYLE_FILL);
-    cairo_fill(mCairo);
+    cairo_fill_preserve(mCairo);
     return Redraw();
 }
 
@@ -1055,14 +1055,14 @@ NS_IMETHODIMP
 nsCanvasRenderingContext2D::Stroke()
 {
     ApplyStyle(STYLE_STROKE);
-    cairo_stroke(mCairo);
+    cairo_stroke_preserve(mCairo);
     return Redraw();
 }
 
 NS_IMETHODIMP
 nsCanvasRenderingContext2D::Clip()
 {
-    cairo_clip(mCairo);
+    cairo_clip_preserve(mCairo);
     return Redraw();
 }
 
@@ -1083,7 +1083,20 @@ nsCanvasRenderingContext2D::LineTo(float x, float y)
 NS_IMETHODIMP
 nsCanvasRenderingContext2D::QuadraticCurveTo(float cpx, float cpy, float x, float y)
 {
-    cairo_curve_to(mCairo, cpx, cpy, cpx, cpy, x, y);
+    double cx, cy;
+
+    cairo_get_current_point(mCairo, &cx, &cy);
+    if (cairo_status(mCairo))
+        return NS_ERROR_INVALID_ARG;
+
+    cairo_curve_to(mCairo,
+                   (cx + cpx * 2.0) / 3.0,
+                   (cy + cpy * 2.0) / 3.0,
+                   (cpx * 2.0 + x) / 3.0,
+                   (cpy * 2.0 + y) / 3.0,
+                   x,
+                   y);
+
     return NS_OK;
 }
 
@@ -1771,6 +1784,29 @@ nsCanvasRenderingContext2D::CairoSurfaceFromElement(nsIDOMElement *imgElt,
     return NS_OK;
 }
 
+static PRBool
+CheckSaneImageSize (PRInt32 width, PRInt32 height)
+{
+    if (width <= 0 || height <= 0)
+        return PR_FALSE;
+
+    /* check to make sure we don't overflow a 32-bit */
+    PRInt32 tmp = width * height;
+    if (tmp / height != width)
+        return PR_FALSE;
+
+    tmp = tmp * 4;
+    if (tmp / 4 != width * height)
+        return PR_FALSE;
+
+    /* reject over-wide or over-tall images */
+    const PRInt32 k64KLimit = 0x0000FFFF;
+    if (width > k64KLimit || height > k64KLimit)
+        return PR_FALSE;
+
+    return PR_TRUE;
+}
+
 NS_IMETHODIMP
 nsCanvasRenderingContext2D::DrawWindow(nsIDOMWindow* aWindow, PRInt32 aX, PRInt32 aY,
                                        PRInt32 aW, PRInt32 aH, 
@@ -1778,6 +1814,11 @@ nsCanvasRenderingContext2D::DrawWindow(nsIDOMWindow* aWindow, PRInt32 aX, PRInt3
 {
     NS_ENSURE_ARG(aWindow != nsnull);
 
+    // protect against too-large surfaces that will cause allocation
+    // or overflow issues
+    if (!CheckSaneImageSize (aW, aH))
+        return NS_ERROR_FAILURE;
+
     // We can't allow web apps to call this until we fix at least the
     // following potential security issues:
     // -- rendering cross-domain IFRAMEs and then extracting the results
@@ -1875,6 +1916,25 @@ nsCanvasRenderingContext2D::DrawWindow(nsIDOMWindow* aWindow, PRInt32 aX, PRInt3
     return rv;
 }
 
+/**
+ * Given aBits, the number of bits in a color channel, compute a number N
+ * such that for values v with aBits bits, floor((N*v)/256) is close to
+ * v*255.0/(2^aBits - 1) and in particular we need
+ * floor((N*(2^aBits - 1))/256) = 255.
+ * We'll just use a table that gives good results :-).
+ */
+static PRUint32 ComputeScaleFactor(PRUint32 aBits)
+{
+  static PRUint32 table[9] = {
+    0, 255*256, 85*256, 9330, 17*256, 2110, 1038, 515, 256
+  };
+  
+  NS_ASSERTION(aBits <= 8, "more than 8 bits in a color channel not supported");
+  NS_ASSERTION(((table[aBits]*((1 << aBits) - 1)) >> 8) == 255,
+               "Invalid table entry");
+  return table[aBits];
+}
+
 nsresult
 nsCanvasRenderingContext2D::DrawNativeSurfaces(nsIDrawingSurface* aBlackSurface,
                                                nsIDrawingSurface* aWhiteSurface,
@@ -1885,7 +1945,12 @@ nsCanvasRenderingContext2D::DrawNativeSurfaces(nsIDrawingSurface* aBlackSurface,
         NS_ERROR("Must have image frame already");
         return NS_ERROR_FAILURE;
     }
-    
+
+    // check if the dimensions are too large;
+    // if they are, we may easily overflow malloc later on
+    if (!CheckSaneImageSize (aSurfaceSize.width, aSurfaceSize.height))
+        return NS_ERROR_FAILURE;
+
     // Acquire alpha values
     nsAutoArrayPtr<PRUint8> alphas;
     nsresult rv;
@@ -2007,6 +2072,11 @@ nsCanvasRenderingContext2D::DrawNativeSurfaces(nsIDrawingSurface* aBlackSurface,
     // Convert the data
     PRUint8* dest = tmpBuf;
     PRInt32 index = 0;
+    
+    PRUint32 RScale = ComputeScaleFactor(format.mRedCount);
+    PRUint32 GScale = ComputeScaleFactor(format.mGreenCount);
+    PRUint32 BScale = ComputeScaleFactor(format.mBlueCount);
+    
     for (PRInt32 i = 0; i < aSurfaceSize.height; ++i) {
         PRUint8* src = data + i*rowSpan;
         for (PRInt32 j = 0; j < aSurfaceSize.width; ++j) {
@@ -2020,12 +2090,12 @@ nsCanvasRenderingContext2D::DrawNativeSurfaces(nsIDrawingSurface* aBlackSurface,
             // Note that because aBlackSurface is the image rendered
             // onto black, the channel values we get here have
             // effectively been premultipled by the alpha value.
-            dest[BLUE_BYTE] = (PRUint8)(((v & format.mBlueMask) >> format.mBlueShift)
-                                << (8 - format.mBlueCount));
-            dest[GREEN_BYTE] = (PRUint8)(((v & format.mGreenMask) >> format.mGreenShift)
-                                << (8 - format.mGreenCount));
-            dest[RED_BYTE] = (PRUint8)(((v & format.mRedMask) >> format.mRedShift)
-                                << (8 - format.mRedCount));
+            dest[BLUE_BYTE] = 
+              (PRUint8)((((v & format.mBlueMask) >> format.mBlueShift)*BScale) >> 8);
+            dest[GREEN_BYTE] =
+              (PRUint8)((((v & format.mGreenMask) >> format.mGreenShift)*GScale) >> 8);
+            dest[RED_BYTE] =
+              (PRUint8)((((v & format.mRedMask) >> format.mRedShift)*RScale) >> 8);
             dest[ALPHA_BYTE] = alphas ? alphas[index++] : 0xFF;
             src += bytesPerPix;
             dest += 4;
