@@ -45,8 +45,10 @@
 #include "nsIDOMNSXPathExpression.h"
 #include "nsIDOMXPathResult.h"
 #include "nsDeque.h"
-#include "nsIModelElementPrivate.h"
 #include "nsXFormsUtils.h"
+#include "nsDOMError.h"
+#include "nsIDOMElement.h"
+#include "nsXFormsModelElement.h"
 
 #ifdef DEBUG
 //#  define DEBUG_XF_MDG
@@ -63,7 +65,6 @@
 /* ------------------------------------ */
 /* --------- nsXFormsMDGNode ---------- */
 /* ------------------------------------ */
-MOZ_DECL_CTOR_COUNTER(nsXFormsMDGNode)
 
 nsXFormsMDGNode::nsXFormsMDGNode(nsIDOMNode             *aNode,
                                  const ModelItemPropName aType)
@@ -121,7 +122,6 @@ nsXFormsMDGNode::MarkClean()
 /* ------------------------------------ */
 /* -------- nsXFormsMDGEngine --------- */
 /* ------------------------------------ */
-MOZ_DECL_CTOR_COUNTER(nsXFormsMDGEngine)
 
 nsXFormsMDGEngine::nsXFormsMDGEngine()
 : mNodesInGraph(0)
@@ -137,7 +137,7 @@ nsXFormsMDGEngine::~nsXFormsMDGEngine()
 }
 
 nsresult
-nsXFormsMDGEngine::Init(nsIModelElementPrivate *aModel)
+nsXFormsMDGEngine::Init(nsXFormsModelElement *aModel)
 {
   nsresult rv = NS_ERROR_FAILURE;
   if (mNodeStates.Init() && mNodeToMDG.Init()) {
@@ -164,7 +164,7 @@ nsXFormsMDGEngine::AddMIP(ModelItemPropName         aType,
   nsAutoString nodename;
   aContextNode->GetNodeName(nodename);
   printf("nsXFormsMDGEngine::AddMIP(aContextNode=%s, aExpression=%p, aDependencies=|%d|,\n",
-         NS_ConvertUCS2toUTF8(nodename).get(),
+         NS_ConvertUTF16toUTF8(nodename).get(),
          (void*) aExpression,
          aDependencies ? aDependencies->Count() : 0);
   printf("                          aContextPos=%d, aContextSize=%d, aType=%s, aDynFunc=%d)\n",
@@ -274,7 +274,7 @@ nsXFormsMDGEngine::PrintDot(const char* aFile)
       
       if (g->IsDirty()) {
         fprintf(FD, "\t%s [color=red];\n",
-                NS_ConvertUCS2toUTF8(domNodeName).get());
+                NS_ConvertUTF16toUTF8(domNodeName).get());
       }
 
       for (PRInt32 j = 0; j < g->mSuc.Count(); ++j) {
@@ -284,8 +284,8 @@ nsXFormsMDGEngine::PrintDot(const char* aFile)
           nsAutoString sucName;
           sucnode->mContextNode->GetNodeName(sucName);
           fprintf(FD, "\t%s -> %s [label=\"%s\"];\n",
-                  NS_ConvertUCS2toUTF8(sucName).get(),
-                  NS_ConvertUCS2toUTF8(domNodeName).get(),
+                  NS_ConvertUTF16toUTF8(sucName).get(),
+                  NS_ConvertUTF16toUTF8(domNodeName).get(),
                   gMIPNames[sucnode->mType]);
         }
       }
@@ -344,7 +344,7 @@ nsXFormsMDGEngine::Recalculate(nsCOMArray<nsIDOMNode> *aChangedNodes)
     printf("\tNode #%d: This=%p, Dirty=%d, DynFunc=%d, Type=%d, Count=%d, Suc=%d, CSize=%d, CPos=%d, Next=%p, domnode=%s\n",
            i, (void*) g, g->IsDirty(), g->mDynFunc, g->mType,
            g->mCount, g->mSuc.Count(), g->mContextSize, g->mContextPosition,
-           (void*) g->mNext, NS_ConvertUCS2toUTF8(domNodeName).get());
+           (void*) g->mNext, NS_ConvertUTF16toUTF8(domNodeName).get());
 #endif
 
     // Ignore node if it is not dirty
@@ -556,7 +556,11 @@ nsXFormsMDGEngine::Rebuild()
 #endif
 
   if (mGraph.Count() != mNodesInGraph) {
-    NS_WARNING("XForms: There are loops in the MDG\n");
+    nsCOMPtr<nsIDOMElement> modelElement;
+    if (mModel) {
+      modelElement = mModel->GetDOMElement();
+    }
+    nsXFormsUtils::ReportError(NS_LITERAL_STRING("MDGLoopError"), modelElement);
     rv = NS_ERROR_ABORT;
   }
 
@@ -743,6 +747,147 @@ nsXFormsMDGEngine::SetNodeValueInternal(nsIDOMNode       *aContextNode,
   return NS_OK;
 }
 
+nsresult
+nsXFormsMDGEngine::SetNodeContent(nsIDOMNode       *aContextNode,
+                                  nsIDOMNode       *aContentEnvelope,
+                                  PRBool           *aNodeChanged)
+{
+  NS_ENSURE_ARG(aContextNode);
+  NS_ENSURE_ARG(aContentEnvelope);
+
+  // ok, this is tricky.  This function will REPLACE the contents of
+  // aContextNode with the a clone of the contents of aContentEnvelope.  If
+  // aContentEnvelope has no contents, then any contents that aContextNode
+  // has will still be removed.  
+
+  if (aNodeChanged) {
+    *aNodeChanged = PR_FALSE;
+  }
+
+  const nsXFormsNodeState* ns = GetNodeState(aContextNode);
+  NS_ENSURE_TRUE(ns, NS_ERROR_FAILURE);
+
+  // If the node is read-only and not set by a @calculate MIP,
+  // ignore the call
+  if (ns->IsReadonly()) {
+    ///
+    /// @todo Better feedback for readonly nodes? (XXX)
+    return NS_OK;
+  }
+
+  PRUint16 nodeType;
+  nsresult rv = aContextNode->GetNodeType(&nodeType);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  if (nodeType != nsIDOMNode::ELEMENT_NODE) {
+    // got to return something pretty unique that we can check down the road in
+    // order to dispatch any error events
+    return NS_ERROR_DOM_WRONG_TYPE_ERR;
+  }
+
+  // Need to determine if the contents of the context node and content envelope
+  // are already the same.  If so, we can avoid some unnecessary work.
+
+  PRBool hasChildren1, hasChildren2, contentsEqual = PR_FALSE;
+  nsresult rv1 = aContextNode->HasChildNodes(&hasChildren1);
+  nsresult rv2 = aContentEnvelope->HasChildNodes(&hasChildren2);
+  if (NS_SUCCEEDED(rv1) && NS_SUCCEEDED(rv2) && hasChildren1 == hasChildren2) {
+    // First test passed.  Both have the same number of children nodes.
+    if (hasChildren1) {
+      nsCOMPtr<nsIDOMNodeList> children1, children2;
+      PRUint32 childrenLength1, childrenLength2;
+  
+      rv1 = aContextNode->GetChildNodes(getter_AddRefs(children1));
+      rv2 = aContentEnvelope->GetChildNodes(getter_AddRefs(children2));
+
+      if (NS_SUCCEEDED(rv1) && NS_SUCCEEDED(rv2) && children1 && children2) {
+
+        // Both have child nodes.
+        rv1 = children1->GetLength(&childrenLength1);
+        rv2 = children2->GetLength(&childrenLength2);
+        if (NS_SUCCEEDED(rv1) && NS_SUCCEEDED(rv2) && 
+            (childrenLength1 == childrenLength2)) {
+
+          // both have the same number of child nodes.  Now checking to see if
+          // each of the children are equal.
+          for (PRUint32 i = 0; i < childrenLength1; ++i) {
+            nsCOMPtr<nsIDOMNode> child1, child2;
+      
+            rv1 = children1->Item(i, getter_AddRefs(child1));
+            rv2 = children2->Item(i, getter_AddRefs(child2));
+            if (NS_FAILED(rv1) || NS_FAILED(rv2)) {
+              // Unexpected error.  Not as many children in the list as we
+              // were told.
+              return NS_ERROR_UNEXPECTED;
+            }
+      
+            contentsEqual = nsXFormsUtils::AreNodesEqual(child1, child2, PR_TRUE);
+            if (!contentsEqual) {
+              break;
+            }
+          }
+        }
+      }
+    } else {
+      // neither have children
+      contentsEqual = PR_TRUE;
+    }
+  }
+
+  if (contentsEqual) {
+    return NS_OK;
+  }
+
+  // remove any child nodes that aContextNode already contains
+  nsCOMPtr<nsIDOMNode> resultNode;
+  nsCOMPtr<nsIDOMNodeList> childList;
+  rv = aContextNode->GetChildNodes(getter_AddRefs(childList));
+  NS_ENSURE_SUCCESS(rv, rv);
+  if (childList) {
+    PRUint32 length;
+    rv = childList->GetLength(&length);
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    for (PRInt32 i = length-1; i >= 0; i--) {
+      nsCOMPtr<nsIDOMNode> childNode;
+      rv = childList->Item(i, getter_AddRefs(childNode));
+      NS_ENSURE_SUCCESS(rv, rv);
+
+      rv = aContextNode->RemoveChild(childNode, getter_AddRefs(resultNode));
+      NS_ENSURE_SUCCESS(rv, rv);
+    }
+  }
+
+  // add contents of the envelope under aContextNode
+  nsCOMPtr<nsIDOMNode> childNode;
+  rv = aContentEnvelope->GetFirstChild(getter_AddRefs(childNode));
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  nsCOMPtr<nsIDOMDocument> document;
+  rv = aContextNode->GetOwnerDocument(getter_AddRefs(document));
+  NS_ENSURE_STATE(document);
+
+  while (childNode) {
+    nsCOMPtr<nsIDOMNode> importedNode;
+    rv = document->ImportNode(childNode, PR_TRUE, getter_AddRefs(importedNode));
+    NS_ENSURE_STATE(importedNode);
+    rv = aContextNode->AppendChild(importedNode, getter_AddRefs(resultNode));
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    rv = childNode->GetNextSibling(getter_AddRefs(resultNode));
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    resultNode.swap(childNode);
+  }
+
+  // NB: Never reached for Readonly nodes.  
+  if (aNodeChanged) {
+    *aNodeChanged = PR_TRUE;
+  }
+
+  return NS_OK;
+}
+
 const nsXFormsNodeState*
 nsXFormsMDGEngine::GetNodeState(nsIDOMNode *aContextNode)
 {
